// AffineTransformations.cpp : This file contains the 'main' function. Program execution begins and ends there.
//

#include <opencv2/core.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/highgui.hpp>

#include <iostream>

cv::Mat createTransformation(float angle, float tx, float ty, float sx, float sy, float skew_factor, float persp1, float persp2)
{
	// Rotation
	cv::Mat R = cv::Mat::eye(3, 3, CV_32F);
	R.at<float>(0, 0) = cos(angle);
	R.at<float>(0, 1) = -sin(angle);
	R.at<float>(1, 0) = sin(angle);
	R.at<float>(1, 1) = cos(angle);

	// Translation
	cv::Mat t = cv::Mat::eye(3, 3, CV_32F);
	t.at<float>(0, 2) = tx;
	t.at<float>(1, 2) = ty;

	// Scaling
	cv::Mat Scale = cv::Mat::eye(3, 3, CV_32F);
	Scale.at<float>(0, 0) = sx;
	Scale.at<float>(1, 1) = sy;

	// Skewing
	cv::Mat Skew = cv::Mat::eye(3, 3, CV_32F);
	Skew.at<float>(0, 1) = skew_factor;

	cv::Mat T = t * R * Skew * Scale;

	// Perspective transformation
	T.at<float>(2, 0) = persp1;
	T.at<float>(2, 1) = persp2;

	return T;
}

cv::Mat applyTransformation(cv::Mat& img, cv::Mat T, bool isPerspective)
{
	cv::Mat newImg = cv::Mat::zeros(img.size(), CV_8UC3);
	int HEIGHT = img.size().height;
	int WIDTH = img.size().width;

	for (int i = 0; i < HEIGHT; ++i)
	{
		for (int j = 0; j < WIDTH; ++j)
		{
			cv::Mat p(3, 1, CV_32F);
			p.at<float>(0, 0) = i;
			p.at<float>(1, 0) = j;
			p.at<float>(2, 0) = 1.f;

			cv::Mat newP = T * p;
			float newZ = 1.f;
			if (isPerspective) // In case of perspective transformation, homogeneous division needed!
			{
				newZ = newP.at<float>(2, 0);
			}

			int newX = newP.at<float>(0, 0) / newZ;
			int newY = newP.at<float>(1, 0) / newZ;

			if (0 <= newX && newX < HEIGHT && 0 <= newY && newY < WIDTH)
			{
				newImg.at<cv::Vec3b>(newX, newY) = img.at<cv::Vec3b>(i, j);
			}
		}
	}

	return newImg;
}

int main()
{
	cv::Mat img = cv::imread("T:\\opencv\\sources\\samples\\data\\apple.jpg");

	if (img.empty())
	{
		std::cout << "Couldn't read image!" << std::endl;
		return 0;
	}

	cv::Mat T = createTransformation(0.81, 200, 150, 0.5, 0.5, 0.2, 0.0001, 0.002);
	std::cout << "T = " << T << std::endl;
	cv::Mat newImg = applyTransformation(img, T, true);

	cv::imshow("Transformations", newImg);

	int key = cv::waitKey(0);

	return 0;
}

// Run program: Ctrl + F5 or Debug > Start Without Debugging menu
// Debug program: F5 or Debug > Start Debugging menu

// Tips for Getting Started: 
//   1. Use the Solution Explorer window to add/manage files
//   2. Use the Team Explorer window to connect to source control
//   3. Use the Output window to see build output and other messages
//   4. Use the Error List window to view errors
//   5. Go to Project > Add New Item to create new code files, or Project > Add Existing Item to add existing code files to the project
//   6. In the future, to open this project again, go to File > Open > Project and select the .sln file
